import json
from .messages import Config

class ValidationError(Exception):
    def __init__(self, message='', error_code=None, **extra):
        language = Config.get_language()
        msg = message if error_code not in language else language[error_code]
        super().__init__(msg.format(**extra))


class InvalidSchema(Exception):
    def __init__(self, errors: dict):
        self.errors = errors
        super().__init__("Invalid schema: " + json.dumps(errors, default=str, ensure_ascii=False))


class Messages:
    MISSING_REQUIRED_FIELD = 1
    MUST_BE_STRING = 2
    MUST_BE_INTEGER = 3
    MUST_BE_BOOLEAN = 4
    MUST_BE_NUMBER = 5
    MUST_BE_LIST = 6
    MUST_BE_DICT = 7
    MINIMUM_LENGTH = 8
    MAXIMUM_LENGTH = 9
    MUST_BE_EMAIL = 10

EnglishLanguage = {
    Messages.MISSING_REQUIRED_FIELD: "Missing required field",
    Messages.MUST_BE_STRING: "Must be a string",
    Messages.MUST_BE_INTEGER: "Must be a valid integer",
    Messages.MUST_BE_NUMBER: "Must be a valid number",
    Messages.MUST_BE_LIST: "Must be a list",
    Messages.MUST_BE_DICT: "Must be a map",
    Messages.MINIMUM_LENGTH: "At least {minimum_value} character(s).",
    Messages.MAXIMUM_LENGTH: "Maximum of {maximum_value} character(s).",
    Messages.MUST_BE_EMAIL: 'Must be a valid email address',
}


class Config(object):
    __instance = None
    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super(Config,cls).__new__(cls)
            cls.__instance.__initialized = False
        return cls.__instance

    def __init__(self, language=EnglishLanguage):      
        if(self.__initialized): return
        self.__initialized = True
        self.language = language

    @classmethod
    def get_language(cls):
        return cls().language




from .fields import Field
from .exceptions import ValidationError, InvalidSchema
from .messages import Messages
from collections import defaultdict
import inspect
import json

class SchemaCreator(type):
    def __new__(mcs, *args, **kwargs):
        args[2]['_fields'] = list(filter(lambda f: isinstance(f[1], Field) and not f[0].startswith('_'), args[2].items()))
        return type.__new__(mcs, *args)


class Schema(metaclass=SchemaCreator):
    @classmethod
    def from_dict(cls, values, skip_none=False, skip_load=False, skip_validate=False):
        errors = defaultdict(list)
        field_values = {}
        instance = cls()

        for name, field in cls._fields:
            if name in values and values[name] is not None:
                if skip_load:
                    field_values[name] = values[name]
                else:
                    try:
                        value = field.load(values[name])
                        validation_errors = []
                        if not skip_validate:
                            validation_errors = field.validate(value)
                        if validation_errors:
                            errors[name].extend(validation_errors)
                        else:
                            field_values[name] = value
                    except ValidationError as error:
                        errors[name].append(error)
            elif name not in values or (skip_none and values[name] is None):
                if field.default is not None:
                    field_values[name] = field.default
                elif field.required:
                    errors[name].append(ValidationError(error_code=Messages.MISSING_REQUIRED_FIELD))
            elif values[name] is None:
                field_values[name] = None
        if errors:
            raise InvalidSchema(errors)
        
        post_context = cls.post_load(instance, field_values)

        for key, value in post_context.items():
            setattr(instance, key, value)

        return instance

    def post_load(self, context):
        return context

    def to_dict(self, skip_none=False):
        attributes = self.__dict__

        if skip_none:
            return {k:v for k,v in attributes.items() if v is not None}
        return attributes
    
    def merge(self, values, skip_none=False):
        if skip_none:
            for k, v in values.items():
                if v is not None and hasattr(self, k):
                    setattr(self, k, v)
        else:
            for k, v in values.items():
                if hasattr(self, k):
                    setattr(self, k, v)

        return self

    def __str__(self):
        return json.dumps(self.to_dict(skip_none=True))


from .messages import Messages
from .exceptions import ValidationError
import math


class Field:
    def __init__(self, default=None, required=False, validators=[]):
        self.default = default
        self.required = required
        self.validators = validators

    def validate(self, value):
        errors = []
        for validator in self.validators:
            try:
                validator(value)
            except ValidationError as error:
                errors.append(error)
        return errors
    
    def load(self, value):
        return value
        
class String(Field):
    def load(self, value):
        if isinstance(value, str):
            return value
        return str(value)

class Bool(Field):
    def load(self, value):
        if isinstance(value, bool):
            return value
        raise ValidationError(error_code=Messages.MUST_BE_BOOLEAN)

class Int(Field):
    def load(self, value):
        if isinstance(value, int):
            return value
        try:
            return int(value)
        except ValueError:
            raise ValidationError(error_code=Messages.MUST_BE_INTEGER)

class Number(Field):
    def load(self, value):
        if isinstance(value, float):
            return value
        try:
            return float(value)
        except ValueError:
            raise ValidationError(error_code=Messages.MUST_BE_NUMBER)


from .exceptions import ValidationError
from .messages import Messages
import math
import re


class Length:
    def __init__(self, min: int=0, max: int=math.inf):
        self.min = min
        self.max = max

    def __call__(self, value):
        size = len(value)
        if size < self.min:
            raise ValidationError(error_code=Messages.MINIMUM_LENGTH, minimum_value=self.min)
        if size > self.max:
            raise ValidationError(error_code=Messages.MAXIMUM_LENGTH, maximum_value=self.max)


class Email:
    def __call__(self, value):
        if not re.match(r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)', value):
            raise ValidationError(error_code=Messages.MUST_BE_EMAIL)


import setuptools
setuptools.setup(
    name="gesiel",
    version="0.1.8",
    url="https://github.com/gabrielweich/gesiel",
    author="Gabriel Weich",
    author_email="gabrielweich.dev@gmail.com",
    description="Simple and flexible schema definition.",
    long_description=open('README.md').read(),
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=[],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)
